// Fragments - torrent_sorter.rs
// Copyright (C) 2022  Felix Häcker <haeckerfelix@gnome.org>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

use std::cell::RefCell;

use glib::subclass::prelude::*;
use glib::{Enum, ParamSpec, ParamSpecBoolean, ParamSpecEnum, ToValue};
use gtk::glib;
use gtk::prelude::*;
use gtk::subclass::sorter::SorterImpl;
use once_cell::sync::Lazy;
use transmission_gobject::TrTorrent;

mod imp {
    use super::*;

    #[derive(Debug, Default)]
    pub struct FrgTorrentSorter {
        pub descending: RefCell<bool>,
        pub sorting: RefCell<FrgTorrentSorting>,
    }

    #[glib::object_subclass]
    impl ObjectSubclass for FrgTorrentSorter {
        const NAME: &'static str = "FrgTorrentSorter";
        type Type = super::FrgTorrentSorter;
        type ParentType = gtk::Sorter;
    }

    impl ObjectImpl for FrgTorrentSorter {
        fn properties() -> &'static [ParamSpec] {
            static PROPERTIES: Lazy<Vec<ParamSpec>> = Lazy::new(|| {
                vec![
                    ParamSpecBoolean::new("descending", "", "", false, glib::ParamFlags::READWRITE),
                    ParamSpecEnum::new(
                        "sorting",
                        "",
                        "",
                        FrgTorrentSorting::static_type(),
                        FrgTorrentSorting::default() as i32,
                        glib::ParamFlags::READWRITE,
                    ),
                ]
            });

            PROPERTIES.as_ref()
        }

        fn property(&self, obj: &Self::Type, _id: usize, pspec: &ParamSpec) -> glib::Value {
            match pspec.name() {
                "descending" => obj.descending().to_value(),
                "sorting" => obj.sorting().to_value(),
                _ => unimplemented!(),
            }
        }

        fn set_property(
            &self,
            obj: &Self::Type,
            _id: usize,
            value: &glib::Value,
            pspec: &ParamSpec,
        ) {
            match pspec.name() {
                "descending" => obj.set_descending(value.get().unwrap()),
                "sorting" => obj.set_sorting(value.get().unwrap()),
                _ => unimplemented!(),
            }
        }
    }

    impl SorterImpl for FrgTorrentSorter {
        fn order(&self, _sorter: &Self::Type) -> gtk::SorterOrder {
            gtk::SorterOrder::Total
        }

        fn compare(
            &self,
            _sorter: &Self::Type,
            item1: &glib::Object,
            item2: &glib::Object,
        ) -> gtk::Ordering {
            let a = item1.downcast_ref::<TrTorrent>().unwrap();
            let b = item2.downcast_ref::<TrTorrent>().unwrap();
            super::FrgTorrentSorter::torrent_cmp(
                a,
                b,
                *self.sorting.borrow(),
                *self.descending.borrow(),
            )
            .into()
        }
    }
}

glib::wrapper! {
    pub struct FrgTorrentSorter(ObjectSubclass<imp::FrgTorrentSorter>) @extends gtk::Sorter;
}

impl FrgTorrentSorter {
    pub fn new() -> Self {
        Self::default()
    }

    pub fn sorting(&self) -> FrgTorrentSorting {
        *self.imp().sorting.borrow()
    }

    pub fn set_sorting(&self, sorting: FrgTorrentSorting) {
        *self.imp().sorting.borrow_mut() = sorting;
        self.changed(gtk::SorterChange::Different);
    }

    pub fn descending(&self) -> bool {
        *self.imp().descending.borrow()
    }

    pub fn set_descending(&self, descending: bool) {
        *self.imp().descending.borrow_mut() = descending;
        self.changed(gtk::SorterChange::Different);
    }

    fn torrent_cmp(
        a: &TrTorrent,
        b: &TrTorrent,
        sorting: FrgTorrentSorting,
        descending: bool,
    ) -> std::cmp::Ordering {
        let mut torrent_a = a.clone();
        let mut torrent_b = b.clone();

        if descending {
            std::mem::swap(&mut torrent_a, &mut torrent_b);
        }

        match sorting {
            FrgTorrentSorting::Default => std::cmp::Ordering::Equal,
            FrgTorrentSorting::Name => torrent_a.name().cmp(&torrent_b.name()),
            FrgTorrentSorting::Queue => torrent_a.queue_position().cmp(&torrent_b.queue_position()),
        }
    }
}

impl Default for FrgTorrentSorter {
    fn default() -> Self {
        glib::Object::new(&[]).unwrap()
    }
}

#[derive(Display, Copy, Debug, Clone, EnumString, PartialEq, Enum, Eq)]
#[repr(u32)]
#[enum_type(name = "FrgTorrentSorting")]
#[derive(Default)]
pub enum FrgTorrentSorting {
    #[default]
    Default,
    Name,
    Queue,
}
