use glib::subclass::prelude::ObjectSubclass;
use glib::subclass::prelude::*;
use glib::{ObjectExt, ParamSpec, ParamSpecInt, ParamSpecObject, StaticType, ToValue};
use once_cell::sync::Lazy;
use transmission_client::SessionStats;

use std::cell::Cell;

use crate::TrSessionStatsDetails;

mod imp {
    use super::*;

    #[derive(Debug, Default)]
    pub struct TrSessionStats {
        pub torrent_count: Cell<i32>,
        pub active_torrent_count: Cell<i32>,
        pub paused_torrent_count: Cell<i32>,
        pub downloaded_torrent_count: Cell<i32>,
        pub download_speed: Cell<i32>,
        pub upload_speed: Cell<i32>,

        pub cumulative_stats: TrSessionStatsDetails,
        pub current_stats: TrSessionStatsDetails,
    }

    #[glib::object_subclass]
    impl ObjectSubclass for TrSessionStats {
        const NAME: &'static str = "TrSessionStats";
        type ParentType = glib::Object;
        type Type = super::TrSessionStats;
    }

    impl ObjectImpl for TrSessionStats {
        fn properties() -> &'static [ParamSpec] {
            static PROPERTIES: Lazy<Vec<ParamSpec>> = Lazy::new(|| {
                vec![
                    ParamSpecInt::new(
                        "torrent-count",
                        "Torrent Count",
                        "Torrent Count",
                        1,
                        std::i32::MAX,
                        1,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt::new(
                        "active-torrent-count",
                        "Active Torrent Count",
                        "Active Torrent Count",
                        1,
                        std::i32::MAX,
                        1,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt::new(
                        "paused-torrent-count",
                        "Paused Torrent Count",
                        "Paused Torrent Count",
                        1,
                        std::i32::MAX,
                        1,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt::new(
                        "downloaded-torrent-count",
                        "Downloaded Torrent Count",
                        "Downloaded Torrent Count",
                        1,
                        std::i32::MAX,
                        1,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt::new(
                        "download-speed",
                        "Download Speed",
                        "Download Speed",
                        1,
                        std::i32::MAX,
                        1,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecInt::new(
                        "upload-speed",
                        "Upload Speed",
                        "Upload Speed",
                        1,
                        std::i32::MAX,
                        1,
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecObject::new(
                        "cumulative-stats",
                        "Cumulative Stats",
                        "Cumulative Stats",
                        TrSessionStatsDetails::static_type(),
                        glib::ParamFlags::READABLE,
                    ),
                    ParamSpecObject::new(
                        "current-stats",
                        "Current Stats",
                        "Current Stats",
                        TrSessionStatsDetails::static_type(),
                        glib::ParamFlags::READABLE,
                    ),
                ]
            });
            PROPERTIES.as_ref()
        }

        fn property(&self, _obj: &Self::Type, _id: usize, pspec: &ParamSpec) -> glib::Value {
            match pspec.name() {
                "torrent-count" => self.torrent_count.get().to_value(),
                "active-torrent-count" => self.active_torrent_count.get().to_value(),
                "paused-torrent-count" => self.paused_torrent_count.get().to_value(),
                "downloaded-torrent-count" => self.downloaded_torrent_count.get().to_value(),
                "download-speed" => self.download_speed.get().to_value(),
                "upload-speed" => self.upload_speed.get().to_value(),
                "cumulative-stats" => self.cumulative_stats.to_value(),
                "current-stats" => self.current_stats.to_value(),
                _ => unimplemented!(),
            }
        }
    }
}

glib::wrapper! {
    pub struct TrSessionStats(ObjectSubclass<imp::TrSessionStats>);
}

impl TrSessionStats {
    pub fn torrent_count(&self) -> i32 {
        self.property("torrent-count")
    }

    pub fn active_torrent_count(&self) -> i32 {
        self.property("active-torrent-count")
    }

    pub fn paused_torrent_count(&self) -> i32 {
        self.property("paused-torrent-count")
    }

    pub fn downloaded_torrent_count(&self) -> i32 {
        self.property("downloaded-torrent-count")
    }

    pub fn download_speed(&self) -> i32 {
        self.property("download-speed")
    }

    pub fn upload_speed(&self) -> i32 {
        self.property("upload-speed")
    }

    pub fn cumulative_stats(&self) -> TrSessionStatsDetails {
        self.property("cumulative-stats")
    }

    pub fn current_stats(&self) -> TrSessionStatsDetails {
        self.property("current-stats")
    }

    pub(crate) fn refresh_values(&self, rpc_stats: SessionStats, download_count: i32) {
        let imp = imp::TrSessionStats::from_instance(self);

        // torrent_count
        if imp.torrent_count.get() != rpc_stats.torrent_count {
            imp.torrent_count.set(rpc_stats.torrent_count);
            self.notify("torrent-count");
        }

        // active_torrent_count
        if imp.active_torrent_count.get() != rpc_stats.active_torrent_count {
            imp.active_torrent_count.set(rpc_stats.active_torrent_count);
            self.notify("active-torrent-count");
        }

        // paused_torrent_count
        if imp.paused_torrent_count.get() != rpc_stats.paused_torrent_count {
            imp.paused_torrent_count.set(rpc_stats.paused_torrent_count);
            self.notify("paused-torrent-count");
        }

        // downloaded_torrent_count
        if imp.downloaded_torrent_count.get() != download_count {
            imp.downloaded_torrent_count.set(download_count);
            self.notify("downloaded-torrent-count");
        }

        // download_speed
        if imp.download_speed.get() != rpc_stats.download_speed {
            imp.download_speed.set(rpc_stats.download_speed);
            self.notify("download-speed");
        }

        // upload_speed
        if imp.upload_speed.get() != rpc_stats.upload_speed {
            imp.upload_speed.set(rpc_stats.upload_speed);
            self.notify("upload_speed");
        }

        imp.cumulative_stats
            .refresh_values(rpc_stats.cumulative_stats);
        imp.current_stats.refresh_values(rpc_stats.current_stats);
    }
}

impl Default for TrSessionStats {
    fn default() -> Self {
        glib::Object::new(&[]).unwrap()
    }
}
